ROUTE_PREFIX = "/api/v1"

ANALYZER_PREFIX = ROUTE_PREFIX + "/analyzer"
HYBRID_PREFIX = ROUTE_PREFIX + "/hybrid"
EXECUTE_PREFIX = ROUTE_PREFIX + "/execute"
COMBINATORIAL_OPTIMIZATION_PREFIX = ROUTE_PREFIX + "/combinatorial_optimization"

TASKS_SUFFIX = "/tasks"

TASKS_GENERATE_SUFFIX = "/tasks/generate"
TASKS_SOLVE_SUFFIX = "/tasks/solve"

ANALYZER_FULL_PATH = ANALYZER_PREFIX + TASKS_SUFFIX
HYBRID_FULL_PATH = HYBRID_PREFIX + TASKS_SUFFIX
EXECUTE_FULL_PATH = EXECUTE_PREFIX + TASKS_SUFFIX
COMBINATORIAL_OPTIMIZATION_GENERATE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_GENERATE_SUFFIX
)
COMBINATORIAL_OPTIMIZATION_SOLVE_FULL_PATH = (
    COMBINATORIAL_OPTIMIZATION_PREFIX + TASKS_SOLVE_SUFFIX
)
